"""GRAPHISOFT"""

import argparse
from urllib.request import Request
from urllib.error import URLError
from typing import Optional, List
from .commands import BasicCommands
from .versioning import _Versioning


def create_request(port: int) -> Request:
    """Creates request
    """
    assert port in ACConnection._port_range()
    address = 'http://127.0.0.1:' + str(port)
    req = Request(address)
    req.add_header('Content-Type', 'application/json')
    return req


class ACConnection:
    """Represents a living connection to an existing ARCHICAD instance
    
    Attributes:
        version (:obj:`int`): The version of the current ARCHICAD instance.
        build (:obj:`int`): The build number of the current ARCHICAD instance.
        lang (:obj:`str`): The language code of the current ARCHICAD instance.
        commands: The commands can be called through this object. 
        types: The data types of commands can be access through this object.
        utilities: Some other useful commands can be called through this object. 
    """
    __slots__ = ('port', 'request', 'version', 'build', 'lang', 'commands', 'types', 'utilities')

    def __init__(self, port: int):
        self.port = port
        self.request = create_request(port)
        basic_commands = BasicCommands(self.request)
        self.version, self.build, self.lang = basic_commands.GetProductInfo()
        V = _Versioning(self.version, self.build, self.request)
        self.commands = V.commands
        self.types = V.types
        self.utilities = V.utilities

    @staticmethod
    def connect(port: Optional[int] = None):
        """Tries to connect to a running ARCHICAD instance.
        
        Args:
            port (:obj:`int`, optional): The ARCHICAD's port. If not given, the command line or the default port will be used.
        
        Raises:
            ConnectionError: When the connection is unsuccessful.
        
        Returns:
            :obj:`ACConnection`: The living connection.
        """
        conn: Optional[ACConnection] = None
        if port is not None:
            try:
                conn = ACConnection(port)
                if not conn:
                    raise ConnectionError(f"Could not connect to ARCHICAD on port {port}")
            except URLError:
                pass
        else:
            conn = ACConnection.connect_from_args()
            if not conn:
                port = ACConnection.find_first_port()
                if port:
                    conn = ACConnection.connect(port)
        return conn

    @staticmethod
    def connect_from_args():
        port = ACConnection.port_from_args()
        if not port:
            return None

        return ACConnection.connect(port)

    @staticmethod
    def port_from_args() -> Optional[int]:
        """Tries to extract hostname and port number from command line arguments
        Returns the tuple of the hostame and port number or None
        """
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('--port', type=int, default=None)
        args = parser.parse_args()
        return args.port
        
    @staticmethod
    def _port_range() -> range:
        return range(19723, 19744, 1)

    @staticmethod
    def find_first_port() -> int:
        for port in ACConnection._port_range():
            try:
                ACConnection(port)
                return port
            except Exception:
                continue
        return None

    @staticmethod
    def find_ports(ports: slice) -> List[int]:
        result = []
        for port in ACConnection._port_range():
            try:
                ACConnection(port)
                result.append(port)
            except Exception:
                continue
        return result



import unittest
from archicad.validators import value_set, matches, min_length, max_length, multiple_of, minimum, maximum, \
    listitem_validator, min_items, max_items, unique_items

class TestStringValidators(unittest.TestCase):
    def test_matches(self):
        pattern = r'^[1-2][0-9]{3}-(0[1-9]|1[0-2])-((0[1-9])|([1-2][0-9])|(3[0-1]))$'
        test_matches = matches(pattern)
        self.assertTrue(test_matches('1982-01-01'))
        self.assertFalse(test_matches('22-12-2012'))

    def test_min_length(self):
        test_min_length = min_length(4)
        self.assertFalse(test_min_length("ONE"))
        self.assertFalse(test_min_length("TWO"))
        self.assertTrue(test_min_length("THREE"))
        self.assertTrue(test_min_length("FOUR"))

    def test_max_length(self):
        test_max_length = max_length(4)
        self.assertTrue(test_max_length("ONE"))
        self.assertTrue(test_max_length("TWO"))
        self.assertTrue(test_max_length("FOUR"))
        self.assertFalse(test_max_length("THREE"))

    def test_value_set(self):
        test_value_set = value_set(['ONE', 'THREE', 'FOUR'])
        self.assertTrue(test_value_set('ONE'))
        self.assertTrue(test_value_set('THREE'))
        self.assertTrue(test_value_set('FOUR'))
        self.assertFalse(test_value_set('TWO'))


class TestNumberValidators(unittest.TestCase):
    def test_multiple_of(self):
        test_int_multiple_of = multiple_of(2)
        self.assertTrue(test_int_multiple_of(4))
        self.assertTrue(test_int_multiple_of(2))
        self.assertFalse(test_int_multiple_of(3))
        self.assertFalse(test_int_multiple_of(1))
        self.assertTrue(test_int_multiple_of(0))

        test_float_multiple_of = multiple_of(2.5)
        self.assertTrue(test_float_multiple_of(2.5))
        self.assertTrue(test_float_multiple_of(10))
        self.assertFalse(test_float_multiple_of(10.1))
        self.assertFalse(test_float_multiple_of(9.9))
        self.assertFalse(test_float_multiple_of(2.6))
        self.assertFalse(test_float_multiple_of(2.4))
        self.assertTrue(test_float_multiple_of(0.0))

    def test_minimum(self):
        test_int_minimum = minimum(5, False)
        self.assertTrue(test_int_minimum(5))
        self.assertFalse(test_int_minimum(-5))
        self.assertFalse(test_int_minimum(0))
        self.assertFalse(test_int_minimum(4))
        self.assertTrue(test_int_minimum(6))
        self.assertTrue(test_int_minimum(1000))

        test_int_exclusive_minimum = minimum(5, True)
        self.assertFalse(test_int_exclusive_minimum(5))
        self.assertFalse(test_int_exclusive_minimum(-5))
        self.assertFalse(test_int_exclusive_minimum(0))
        self.assertFalse(test_int_exclusive_minimum(4))
        self.assertTrue(test_int_exclusive_minimum(6))
        self.assertTrue(test_int_exclusive_minimum(1000))
        
        test_float_minimum = minimum(5.5, False)
        self.assertTrue(test_float_minimum(5.5))
        self.assertFalse(test_float_minimum(-5.5))
        self.assertFalse(test_float_minimum(0))
        self.assertFalse(test_float_minimum(4))
        self.assertFalse(test_float_minimum(5.4))
        self.assertFalse(test_float_minimum(4.5))
        self.assertTrue(test_float_minimum(5.6))
        self.assertTrue(test_float_minimum(6.6))
        self.assertTrue(test_float_minimum(6))

        test_exclusive_minimum = minimum(5.5, True)
        self.assertFalse(test_exclusive_minimum(5.5))
        self.assertFalse(test_exclusive_minimum(-5.5))
        self.assertFalse(test_exclusive_minimum(0))
        self.assertFalse(test_exclusive_minimum(4))
        self.assertFalse(test_exclusive_minimum(5.4))
        self.assertFalse(test_exclusive_minimum(5.499999))
        self.assertFalse(test_exclusive_minimum(4.5))
        self.assertTrue(test_exclusive_minimum(5.5000001))
        self.assertTrue(test_exclusive_minimum(5.6))
        self.assertTrue(test_exclusive_minimum(6.6))
        self.assertTrue(test_exclusive_minimum(6))

    def test_maximum(self):
        test_int_maximum = maximum(5, False)
        self.assertTrue(test_int_maximum(5))
        self.assertTrue(test_int_maximum(-5))
        self.assertTrue(test_int_maximum(0))
        self.assertTrue(test_int_maximum(4))
        self.assertFalse(test_int_maximum(6))
        self.assertFalse(test_int_maximum(1000))

        test_int_exclusive_maximum = maximum(5, True)
        self.assertFalse(test_int_exclusive_maximum(5))
        self.assertTrue(test_int_exclusive_maximum(-5))
        self.assertTrue(test_int_exclusive_maximum(0))
        self.assertTrue(test_int_exclusive_maximum(4))
        self.assertFalse(test_int_exclusive_maximum(6))
        self.assertFalse(test_int_exclusive_maximum(1000))
        
        test_float_maximum = maximum(5.5, False)
        self.assertTrue(test_float_maximum(5.5))
        self.assertTrue(test_float_maximum(-5.5))
        self.assertTrue(test_float_maximum(0))
        self.assertTrue(test_float_maximum(4))
        self.assertTrue(test_float_maximum(5.4))
        self.assertTrue(test_float_maximum(4.5))
        self.assertFalse(test_float_maximum(5.6))
        self.assertFalse(test_float_maximum(6.6))
        self.assertFalse(test_float_maximum(6))

        test_exclusive_maximum = maximum(5.5, True)
        self.assertFalse(test_exclusive_maximum(5.5))
        self.assertTrue(test_exclusive_maximum(-5.5))
        self.assertTrue(test_exclusive_maximum(0))
        self.assertTrue(test_exclusive_maximum(4))
        self.assertTrue(test_exclusive_maximum(5.4))
        self.assertTrue(test_exclusive_maximum(5.499999))
        self.assertTrue(test_exclusive_maximum(4.5))
        self.assertFalse(test_exclusive_maximum(5.5000001))
        self.assertFalse(test_exclusive_maximum(5.6))
        self.assertFalse(test_exclusive_maximum(6.6))
        self.assertFalse(test_exclusive_maximum(6))
    
    def test_value_set(self):
        test_value_set = value_set([4, 5.5, .5, 3.])
        self.assertTrue(test_value_set(4))
        self.assertTrue(test_value_set(3))
        self.assertTrue(test_value_set(3.))
        self.assertTrue(test_value_set(4.))
        self.assertTrue(test_value_set(5.5))
        self.assertTrue(test_value_set(.5))
        self.assertTrue(test_value_set(0.5))
        self.assertFalse(test_value_set(5.4))
        self.assertFalse(test_value_set(5.6))
        self.assertFalse(test_value_set(3.01))
        self.assertFalse(test_value_set(2.99))
        self.assertFalse(test_value_set(4.0001))
        self.assertFalse(test_value_set(3.99))
        self.assertFalse(test_value_set(.51))
        self.assertFalse(test_value_set(.49))
        self.assertFalse(test_value_set(6))
        self.assertFalse(test_value_set(1))
        self.assertFalse(test_value_set(0))


class TestListValidators(unittest.TestCase):
    def test_min_items(self):
        test_min_items = min_items(3)
        self.assertTrue(test_min_items([1,2,3,4,5,6]))
        self.assertTrue(test_min_items(["1","2","3"]))
        self.assertTrue(test_min_items([1,2,3]))
        self.assertFalse(test_min_items([1,2]))
        self.assertFalse(test_min_items([1,]))
        self.assertFalse(test_min_items(["123456"]))
        self.assertFalse(test_min_items(["12"]))

    def test_max_items(self):
        test_max_items = max_items(3)
        self.assertFalse(test_max_items([1,2,3,4,5,6]))
        self.assertFalse(test_max_items(["1","2","3","4"]))
        self.assertTrue(test_max_items(["1","2","3"]))
        self.assertTrue(test_max_items([1,2,3]))
        self.assertTrue(test_max_items([1,2]))
        self.assertTrue(test_max_items([123456]))
        self.assertTrue(test_max_items(["123456"]))
        self.assertTrue(test_max_items(["12"]))


    def test_unique_items(self):
        test_unique_items = unique_items()
        self.assertTrue(test_unique_items([]))
        self.assertTrue(test_unique_items([1]))
        self.assertTrue(test_unique_items(["1"]))
        self.assertTrue(test_unique_items([1,2,3,4,5,6]))
        self.assertTrue(test_unique_items(["1","2","3","4","5","6"]))
        self.assertFalse(test_unique_items([1,2,2.,3,4,5,6]))
        self.assertFalse(test_unique_items(["1","212","3","4","212","5","6"]))
        self.assertFalse(test_unique_items([1,1,1,1]))

    def test_listitem_validator(self):
        test_listitem_validator = listitem_validator(multiple_of(2), minimum(4, False), maximum(16, True))
        self.assertTrue(test_listitem_validator([4,6,10]))
        self.assertTrue(test_listitem_validator([4,6,6,10,14]))
        self.assertFalse(test_listitem_validator([2,4,6]))
        self.assertFalse(test_listitem_validator([2,4,16]))
        self.assertFalse(test_listitem_validator([2,3,6,11]))




"""The archicad package from GRAPHISOFT
"""

from .handlers import handle_dependencies
from .connection import ACConnection
from .releases import Commands, Types, Utilities

__all__ = ['ACConnection', 'handle_dependencies', 'Commands', 'Types', 'Utilities']
